#include "tff.h"
#include <avr/io.h>
#include <avr/sleep.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include "mmc_hardware.h"
#include "diskio.h"

static  prog_char pat_RIFF[] = "RIFF";
static  prog_char pat_WAVE[] = "WAVE";
static  prog_char pat_FMT[] = "fmt ";

#ifdef _AVR_IOM128_H_

#define LEFT_LO_PORT PORTF
#define LEFT_LO_DDR DDRF
#define LEFT_LO_SHIFT 0
#define LEFT_HI_PORT PORTF
#define LEFT_HI_DDR DDRF
#define LEFT_HI_SHIFT 0

#define RIGHT_LO_PORT PORTB
#define RIGHT_LO_DDR DDRB
#define RIGHT_LO_SHIFT 0
#define RIGHT_HI_PORT PORTB
#define RIGHT_HI_DDR DDRB
#define RIGHT_HI_SHIFT 1

#else

#define LEFT_LO_PORT PORTC
#define LEFT_LO_DDR DDRC
#define LEFT_LO_SHIFT 1
#define LEFT_HI_PORT PORTD
#define LEFT_HI_DDR DDRD
#define LEFT_HI_SHIFT 0

#define RIGHT_LO_PORT PORTB
#define RIGHT_LO_DDR DDRB
#define RIGHT_LO_SHIFT 0
#define RIGHT_HI_PORT PORTA
#define RIGHT_HI_DDR DDRA
#define RIGHT_HI_SHIFT 0

#endif

BYTE nchannels;
volatile WORD dt;
volatile BYTE buf_empty = 0;

FIL file;
FRESULT res;
FATFS work;
BYTE buf[1024];
volatile WORD bufpos = 0;

/*********************************************************/

DWORD get_fattime(void)
{
	return ((2008L-1980) << 25) | (1L << 21) | (1L << 16);
}

BYTE strncmp_f(BYTE *buf, char *pattern, BYTE n)
{
	BYTE i;
	for (i = 0; i < n; i++) {
		if (buf[i] != pgm_read_byte(pattern + i)) return 1;
	}
	return 0;
}

BYTE parse_wav_header(BYTE *header)
{
	DWORD tmp;
	if (strncmp_f(header, pat_RIFF, 4)) return 0;
	if (strncmp_f(header + 8, pat_WAVE, 4)) return 0;
	if (strncmp_f(header + 12, pat_FMT, 4)) return 0;
	if (((DWORD *)(header + 16))[0] != 0x10) return 0;
	nchannels = ((DWORD *)(header + 22))[0];
	if (nchannels != 1 && nchannels != 2) return 0;
	tmp = ((DWORD *)(header + 24))[0];
	if (tmp < (F_CPU / 20000) || tmp > (F_CPU / 250)) return 0;
	dt = F_CPU / tmp;
	tmp = ((WORD *)(header + 32))[0];
	if (nchannels != tmp) return 0;
	return 1;
}

//SIGNAL(SIG_OUTPUT_COMPARE1A) __attribute__ (( __naked__ ));
SIGNAL(SIG_OUTPUT_COMPARE1A)
{
	BYTE left , right;
	OCR1A += dt;
	left = buf[bufpos++];
#if LEFT_LO_SHIFT
	LEFT_LO_PORT = (LEFT_LO_PORT & 0x0F) | left << 4;
#else
	LEFT_LO_PORT = (LEFT_LO_PORT & 0xF0) | (left & 0x0F);
#endif
#if LEFT_HI_SHIFT
	LEFT_HI_PORT = (LEFT_HI_PORT & 0xF0) | left >> 4;
#else
	LEFT_HI_PORT = (LEFT_HI_PORT & 0x0F) | (left & 0xF0);
#endif
	if (nchannels == 1) {
		right = left;
	} else {
		right = buf[bufpos++];
	}
#if RIGHT_LO_SHIFT
	RIGHT_LO_PORT = (RIGHT_LO_PORT & 0x0F) | right << 4;
#else
	RIGHT_LO_PORT = (RIGHT_LO_PORT & 0xF0) | (right & 0x0F);
#endif
#if RIGHT_HI_SHIFT
	RIGHT_HI_PORT = (RIGHT_HI_PORT & 0xF0) | right >> 4;
#else
	RIGHT_HI_PORT = (RIGHT_HI_PORT & 0x0F) | (right & 0xF0);
#endif
	bufpos &= 1023;
}


int main(void)
{
#if RIGHT_LO_SHIFT
	RIGHT_LO_DDR |= 0xF0;
#else
	RIGHT_LO_DDR |= 0x0F;
#endif
#if RIGHT_HI_SHIFT
	RIGHT_HI_DDR |= 0x0F;
#else
	RIGHT_HI_DDR |= 0xF0;
#endif
#if LEFT_LO_SHIFT
	LEFT_LO_DDR |= 0xF0;
#else
	LEFT_LO_DDR |= 0x0F;
#endif
#if LEFT_HI_SHIFT
	LEFT_HI_DDR |= 0x0F;
#else
	LEFT_HI_DDR |= 0xF0;
#endif

	TCCR1B = 1 << CS10;

	res = f_mount(0, &work);
	sei();
	res = f_open (&file, "/mono16.wav", FA_READ);
	if (res == FR_OK) {
		UINT bytes_read;
		f_read(&file, buf, 44, &bytes_read);
		if (parse_wav_header(buf)) {
		} else {
		}
		f_read(&file, buf, 512, &bytes_read);
		buf_empty = 0;
		bufpos = 0;
		TIMSK |= _BV(OCIE1A);
		while (1) {
			WORD pos;
			TIMSK &= ~_BV(OCIE1A);
			pos = bufpos;
			TIMSK |= _BV(OCIE1A);
			if (buf_empty && pos >= 512) {
					buf_empty = 0;
					res = f_read(&file, buf + 512, 512, &bytes_read);
					if (!bytes_read) break;
			} else if (!buf_empty && pos < 512) {
					buf_empty = 1;
					res = f_read(&file, buf, 512, &bytes_read);
					if (!bytes_read) break;
			}
		}
		TIMSK &= ~_BV(OCIE1A);
	}
	for(;;);
}
